/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <gtk/gtk.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gtksourceview/gtksourceview.h>
#include <fcntl.h>
#include "tcpjunk.h"

struct gui_elements
  {
	GtkWidget *window;
	GtkWidget *host_entry;
	GtkWidget *port_entry;
	GtkWidget *sport_entry;
	GtkWidget *session_log_text;
	GtkWidget *connect_butt;
	GtkWidget *listen_butt;
	GtkWidget *status_text;
	GtkWidget *init_to_entry;
	GtkWidget *sock_to_entry;
	GtkWidget *repeat_entry;
	GtkWidget *syntax_entry;
	GtkWidget *ses_tabs;
	GtkWidget *open_mi;
	char srv_addr[FILENAME_MAX];
	char *cur_folder;
	int auto_scroll;
	int auto_clearstat;
	int gui_listen_sock;
	int listening;
	char *undo_buf;
	char *redo_buf;
  };

struct gui_elements gui_data;

static GtkWidget *get_tab_textview()
  {
  gint current_tab_num = 0;
  GtkWidget *curr_frame, *curr_scrwin, *curr_textview;
  current_tab_num = gtk_notebook_get_current_page(GTK_NOTEBOOK(
	  gui_data.ses_tabs));
  curr_frame = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gui_data.ses_tabs),
	  current_tab_num);
  curr_scrwin = gtk_bin_get_child(GTK_BIN(curr_frame));
  curr_textview = gtk_bin_get_child(GTK_BIN(curr_scrwin));
  return curr_textview;
  }

static gboolean exit_gui(GtkWidget *widget, GdkEvent *event, gpointer data)
  {
  gtk_main_quit();
  return FALSE;
  }

static void gui_update()
  {

  while (gtk_events_pending())
	gtk_main_iteration();
  }

static void gui_clear_session_log()
  {
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_log_text));
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  gtk_text_buffer_delete(buffer, &start, &end);
  gui_update();
  }

static void gui_clear_status()
  {
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.status_text));
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  gtk_text_buffer_delete(buffer, &start, &end);
  gui_update();
  }

static void gui_about(struct tj_data *gui_inst_data)
  {
  gui_clear_status();
  char about_str[1024];
  sprintf(
	  about_str,
	  "Tcpjunk version %s\n\
(c) 2008-2009 Ofer Kashayov\n\
http://code.google.com/p/tcpjunk/",
	  VER);
  tj_error(gui_inst_data, about_str, "", c_black);

  }

void gui_insert_tag(char *tag_text)
  {
  GtkTextBuffer *ses_buf;
  GtkTextIter start, end;

  /* get existing text */
  ses_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  gtk_text_buffer_get_bounds(ses_buf, &start, &end);

  /* insert text */
  gtk_text_buffer_insert_at_cursor(ses_buf, tag_text, strlen(tag_text));

  }

void gui_insert_sect()
  {
  char *tag_text = "<sect>user root\n</sect>";
  gui_insert_tag(tag_text);
  }

void gui_insert_rep()
  {
  char *tag_text = "<rep 1 10>text</rep>";
  gui_insert_tag(tag_text);
  }

void gui_insert_sys()
  {
  char *tag_text = "<sys>pwd</sys>";
  gui_insert_tag(tag_text);
  }

void gui_insert_fuzz()
  {
  char *tag_text = "<fuzz alnum 1 10>";
  gui_insert_tag(tag_text);
  }

void gui_insert_char()
  {
  char *tag_text = "<char 5a 10>";
  gui_insert_tag(tag_text);
  }

void gui_insert_enc()
  {
  char *tag_text = "<enc % hex none 2>Hi there!!</enc>";
  gui_insert_tag(tag_text);
  }

void gui_insert_count()
  {
  char *tag_text = "<count hex 1 10>";
  gui_insert_tag(tag_text);
  }

static void gui_scroll_session_log()
  {

  /* auto scroll */
  GtkTextBuffer *buffer;
  GtkTextIter end;

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_log_text));
  gtk_text_buffer_get_end_iter(buffer, &end);
  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(gui_data.session_log_text), &end,
	  0, TRUE, 0, 0);
  }

static void gui_show_selection(GtkWidget *text_view,
	struct tj_data *gui_inst_data)
  {
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_view));
  gtk_text_buffer_get_selection_bounds(buffer, &start, &end);
  char tempbuf[64];
  int chars_total;
  chars_total = gtk_text_iter_get_offset(&end) - gtk_text_iter_get_offset(
	  &start);
  sprintf(tempbuf, "%d characters copied (%x)", chars_total, chars_total);
  tj_error(gui_inst_data, tempbuf, "", c_black);
  }

static void text_buffer_changed(struct tj_data *gui_inst_data,
	GtkTextBuffer *sess_textbuf)
  {
  gui_inst_data->sess_data_modified = 1;
  }

static void new_tab(struct tj_data *gui_inst_data)
  {
  GtkWidget *new_lab, *new_textview, *new_frame, *new_scrwin;
  GtkTextBuffer *new_text_buf;

  /* textview */
  new_textview = gtk_source_view_new();
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(new_textview), GTK_WRAP_CHAR);

  g_signal_connect(GTK_TEXT_VIEW(new_textview), "copy-clipboard", G_CALLBACK(
		  gui_show_selection), gui_inst_data);

  /* textview buf */
  new_text_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(new_textview));
  gtk_text_buffer_create_tag(new_text_buf, "yellow_hl", "background",
	  "#afffaf", NULL);
  g_signal_connect_swapped(G_OBJECT(new_text_buf), "changed", G_CALLBACK(
		  text_buffer_changed), gui_inst_data);

  /* textview fonts */
  PangoFontDescription *font_desc;
  font_desc = pango_font_description_from_string("Fixed 13");
  gtk_widget_modify_font(new_textview, font_desc);

  /* scroll win */
  new_scrwin = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(new_scrwin),
	  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_container_add(GTK_CONTAINER(new_scrwin), new_textview);

  /* frame */
  new_frame = gtk_frame_new(NULL);
  gtk_container_add(GTK_CONTAINER(new_frame), new_scrwin);

  /* new tab */
  new_lab = gtk_label_new("untitled");
  gtk_notebook_append_page(GTK_NOTEBOOK(gui_data.ses_tabs), new_frame, new_lab);
  gtk_widget_show_all(gui_data.ses_tabs);
  gint npages = gtk_notebook_get_n_pages(GTK_NOTEBOOK(gui_data.ses_tabs));
  gtk_notebook_set_current_page(GTK_NOTEBOOK(gui_data.ses_tabs), npages - 1);
  }

static void close_tab(struct tj_data *gui_inst_data)
  {
  gint current_tab_num = 0;
  current_tab_num = gtk_notebook_get_current_page(GTK_NOTEBOOK(
	  gui_data.ses_tabs));
  gint npages = gtk_notebook_get_n_pages(GTK_NOTEBOOK(gui_data.ses_tabs));
  if (npages > 1)
	gtk_notebook_remove_page(GTK_NOTEBOOK(gui_data.ses_tabs), current_tab_num);
  }

static void syntax_highlight(struct tj_data *gui_inst_data)
  {
  GtkTextBuffer *ses_txt_buf, *ses_log_buf;
  GtkTextIter st_start, st_end, sl_start, sl_end;
  GtkTextTag *ses_text_tag, *ses_log_tag;
  GtkTextIter mstart, mend;
  GtkTextTagTable *tag_table;
  gchar *hl_text;
  int matches = 0;
  char match_msg[128];

  hl_text = (gchar *) gtk_entry_get_text(GTK_ENTRY(gui_data.syntax_entry));

  /* clear existing tags */
  ses_txt_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  gtk_text_buffer_get_bounds(ses_txt_buf, &st_start, &st_end);
  gtk_text_buffer_remove_tag_by_name(ses_txt_buf, "yellow_hl", &st_start,
	  &st_end);
  ses_log_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(
	  gui_data.session_log_text));
  gtk_text_buffer_get_bounds(ses_log_buf, &sl_start, &sl_end);
  gtk_text_buffer_remove_tag_by_name(ses_log_buf, "yellow_hl", &sl_start,
	  &sl_end);

  if (strlen(hl_text) < 1)
	return;

  tag_table = gtk_text_buffer_get_tag_table(ses_txt_buf);
  ses_text_tag = gtk_text_tag_table_lookup(tag_table, "yellow_hl");
  tag_table = gtk_text_buffer_get_tag_table(ses_log_buf);
  ses_log_tag = gtk_text_tag_table_lookup(tag_table, "yellow_hl");

  /* session text */
  while (gtk_text_iter_forward_search(&st_start, hl_text, 0, &mstart, &mend,
	  NULL))
	{
	matches++;
	gtk_text_buffer_apply_tag(ses_txt_buf, ses_text_tag, &mstart, &mend);
	st_start = mend;
	}
  if (matches)
	{
	sprintf(match_msg, "%d highlights in session text", matches);
	tj_error(gui_inst_data, match_msg, "", 0);
	matches = 0;
	}

  /* session log */
  while (gtk_text_iter_forward_search(&sl_start, hl_text, 0, &mstart, &mend,
	  NULL))
	{
	matches++;
	gtk_text_buffer_apply_tag(ses_log_buf, ses_log_tag, &mstart, &mend);
	sl_start = mend;
	}
  if (matches)
	{
	sprintf(match_msg, "%d highlights in session log", matches);
	tj_error(gui_inst_data, match_msg, "", c_black);
	}

  gui_update();

  }

static void set_timeouts(struct tj_data *gui_inst_data)
  {
  /* timeouts */
  gui_inst_data->conn_timeout = atoi(gtk_entry_get_text(GTK_ENTRY(
	  gui_data.init_to_entry)));
  gui_inst_data->reply_timeout = atoi(gtk_entry_get_text(GTK_ENTRY(
	  gui_data.sock_to_entry)));
  gui_inst_data->reply_timeout *= 1000;
  }

static void gui_connect(struct tj_data *gui_inst_data)
  {
  char *gui_session_text = NULL;

  /* misc GUI cleanup */
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.connect_butt), FALSE);
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.listen_butt), FALSE);
  gui_clear_session_log();
  if (gui_data.auto_clearstat == 1)
	gui_clear_status();
  else
	tj_error(gui_inst_data, "-", "", c_black);

  /* check address */
  int check_ok = 0;
  char *temp_str;
  temp_str = (gchar *) gtk_entry_get_text(GTK_ENTRY(gui_data.host_entry));
  strcpy(gui_data.srv_addr, temp_str);
  if (!(check_ok = check_address(gui_inst_data, gui_data.srv_addr)))
	goto done;

  /* check port */
  const gchar *port_text = gtk_entry_get_text(GTK_ENTRY(gui_data.port_entry));

  if (strlen(port_text) == 0)
	{
	tj_error(gui_inst_data, "Invalid port", "", c_red);
	goto done;
	}
  else
	{
	gui_inst_data->dest_port = atoi(port_text);
	if (!(check_port(gui_inst_data->dest_port, gui_inst_data)))
	  goto done;
	}

  /* check source port */
  const gchar *sport_text = gtk_entry_get_text(GTK_ENTRY(gui_data.sport_entry));
  if (strlen(sport_text) > 0)
	{
	gui_inst_data->src_port = atoi(gtk_entry_get_text(GTK_ENTRY(
		gui_data.sport_entry)));
	if (!(check_port(gui_inst_data->src_port, gui_inst_data)))
	  goto done;
	}

  /* fix timeouts */
  set_timeouts(gui_inst_data);

  /* repeat */
  gui_inst_data->keepalive = atoi(gtk_entry_get_text(GTK_ENTRY(
	  gui_data.repeat_entry)));

  /* load session data from GUI */
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  gui_session_text = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

  /* check session data */
  gui_inst_data->sess_data_len = strlen(gui_session_text);
  if (!gui_inst_data->sess_data_len)
	{
	tj_error(gui_inst_data, "Session data is empty", "", c_red);
	goto done;
	}
  gui_inst_data->sess_data = tj_realloc(gui_inst_data->sess_data,
	  gui_inst_data->sess_data_len);
  memcpy(gui_inst_data->sess_data, gui_session_text,
	  gui_inst_data->sess_data_len);
  find_sections(gui_inst_data);

  /* connect */
  junk(gui_inst_data);
  syntax_highlight(gui_inst_data);

  /* finish */
  done:

  if (gui_data.auto_scroll == 1)
	{
	gui_scroll_session_log();
	}
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.connect_butt), TRUE);
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.listen_butt), TRUE);

  if (gui_session_text != NULL)
	g_free(gui_session_text);

  return;

  }

static gboolean gui_check_accept(struct tj_data *gui_inst_data)
  {
  if (gui_data.listening == 0)
	return FALSE;

  /* initialize descriptors set */
  int sel_ret;
  fd_set set;
  FD_ZERO (&set);
  FD_SET (gui_data.gui_listen_sock, &set);

  /* initialize timeout */
  struct timeval timeout;
  timeout.tv_sec = 0;
  timeout.tv_usec = 50;

  sel_ret = select(gui_data.gui_listen_sock + 1, &set, NULL, NULL, &timeout);

  if (sel_ret < 1)
	return TRUE;
  else
	{
	gui_accept(gui_inst_data);
	return FALSE;
	}
  }

static gint syntax_entry_key_press(GtkWidget *widget, GdkEventKey *key,
	struct tj_data *gui_inst_data)
  {
  guint keyval;
  keyval = key->keyval;

  switch (keyval)
	{
  case GDK_Return:
  case GDK_KP_Enter:
  case GDK_ISO_Enter:
  case GDK_3270_Enter:
	syntax_highlight(gui_inst_data);

  default:
	return 0;
	}
  }

static void gui_enable(gboolean state)
  {
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.connect_butt), state);
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.open_mi), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.host_entry), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.port_entry), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.init_to_entry), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.sock_to_entry), state);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(get_tab_textview()), state);
  }

static void gui_stop_listening()
  {
  gui_data.listening = 0;
  gtk_button_set_label(GTK_BUTTON(gui_data.listen_butt), "Listen [Ctrl+L]");
  gui_enable(TRUE);

  close(gui_data.gui_listen_sock);
  return;
  }

void gui_accept(struct tj_data *gui_inst_data)
  {
  /* accept */
  gui_clear_session_log();
  if (gui_data.auto_clearstat == 1)
	gui_clear_status();
  unsigned int sin_size = sizeof(struct sockaddr_in);
  struct sockaddr_in rsa;

  gui_inst_data->srv_sock = accept(gui_data.gui_listen_sock,
	  (struct sockaddr *) &rsa, &sin_size);
  if (gui_inst_data->srv_sock < 0)
	{
	tj_error(gui_inst_data, "Failed accept on socket: ", strerror(errno), c_red);
	gui_stop_listening();
	return;
	}

  /* start transfers */
  tj_error(gui_inst_data, "Connection accepted", "", c_black);

  /* stop checking */
  g_timeout_add(500, (GSourceFunc) gui_check_accept, gui_inst_data);

  /* start session */
  sjunk(gui_inst_data);

  if (gui_data.auto_scroll == 1)
	{
	gui_scroll_session_log();
	}
  }

static void gui_listen(struct tj_data *gui_inst_data)
  {
  if (gui_data.listening == 1)
	{
	gui_stop_listening();
	return;
	}

  /* disable GUI */
  gui_enable(FALSE);
  gtk_button_set_label(GTK_BUTTON(gui_data.listen_butt),
	  "Stop listening [Ctrl+L]");

  /* clear GUI */
  gui_clear_session_log();
  if (gui_data.auto_clearstat == 1)
	gui_clear_status();
  int ret;

  /* check port */
  const gchar *port_text = gtk_entry_get_text(GTK_ENTRY(gui_data.port_entry));

  if (strlen(port_text) == 0)
	{
	tj_error(gui_inst_data, "Invalid port", "", c_red);
	goto failure;
	}
  else
	{
	gui_inst_data->dest_port = atoi(port_text);
	if (!(check_port(gui_inst_data->dest_port, gui_inst_data)))
	  goto failure;
	}

  /* set timeouts */
  set_timeouts(gui_inst_data);

  /* keep alive */
  gui_inst_data->keepalive = atoi(gtk_entry_get_text(GTK_ENTRY(
	  gui_data.repeat_entry)));

  /* get session data from GUI*/
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  gtk_text_buffer_get_bounds(buffer, &start, &end);

  char *gui_session_text =
	  gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

  /* check session data */
  gui_inst_data->sess_data_len = strlen(gui_session_text);
  if (!gui_inst_data->sess_data_len)
	{
	tj_error(gui_inst_data, "Session data is empty", "", c_red);
	goto failure;
	}
  gui_inst_data->sess_data = tj_realloc(gui_inst_data->sess_data,
	  gui_inst_data->sess_data_len);
  memcpy(gui_inst_data->sess_data, gui_session_text,
	  gui_inst_data->sess_data_len);
  find_sections(gui_inst_data);

  /* preparations */
  struct sockaddr_in sa;
  memset(&sa, 0, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = INADDR_ANY;
  sa.sin_port = htons(gui_inst_data->dest_port);
  int yes = 1;

  /* create listening socket */
  gui_data.gui_listen_sock = socket(AF_INET, SOCK_STREAM, 0);
  if (gui_data.gui_listen_sock < 0)
	{
	tj_error(gui_inst_data, "Failed to allocate socket: ", strerror(errno),
		c_red);
	goto failure;
	}

  /* set socket reuse */
  ret = setsockopt(gui_data.gui_listen_sock, SOL_SOCKET, SO_REUSEADDR, &yes,
	  sizeof(int));
  if (ret < 0)
	{
	tj_error(gui_inst_data, "Error setting socket options: ", strerror(errno),
		c_red);
	goto failure;
	}

  /* bind socket */
  ret = bind(gui_data.gui_listen_sock, (struct sockaddr *) &sa,
	  sizeof(struct sockaddr));
  if (ret < 0)
	{
	tj_error(gui_inst_data, "Failed to bind socket: ", strerror(errno), c_red);
	goto failure;
	}

  /* listen on socket */
  ret = listen(gui_data.gui_listen_sock, BACKLOG);
  if (ret < 0)
	{
	tj_error(gui_inst_data, "Failed to listen on socket: ", strerror(errno),
		c_red);
	goto failure;
	}

  /* all was OK */
  g_timeout_add(250, (GSourceFunc) gui_check_accept, gui_inst_data);
  gui_data.listening = 1;
  tj_error(gui_inst_data, "Listening...", "", c_black);
  return;

  /* something wrong */
  failure:

  gui_stop_listening();
  return;

  }

void gui_show_traffic(char *buf, int buf_len, int color,
	struct tj_data *gui_inst_data)
  {
  GtkTextBuffer *ses_log_buf;
  GtkTextIter start, end;
  GtkTextTag *fg_tag;
  GtkTextTagTable *tag_table;
  int i;

  /* disable non printable chars */
  if (!gui_inst_data->hexdump)
	for (i = 0; i < buf_len; i++)
	  {
	  if (isprint(buf[i]) < 1 && isspace(buf[i]) < 1)
		buf[i] = '.';
	  }

  /* get existing text */
  ses_log_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(
	  gui_data.session_log_text));
  gtk_text_buffer_get_bounds(ses_log_buf, &start, &end);
  tag_table = gtk_text_buffer_get_tag_table(ses_log_buf);

  /* choose a color */
  if (color == 0)
	fg_tag = gtk_text_tag_table_lookup(tag_table, "red_fg");
  else
	fg_tag = gtk_text_tag_table_lookup(tag_table, "blue_fg");

  /* insert text */
  if (!gui_inst_data->hexdump)
	gtk_text_buffer_insert_with_tags(ses_log_buf, &end, buf, buf_len, fg_tag,
		NULL);
  else
	{
	char *hex_data_buf = hex_dump(buf, buf_len);
	gtk_text_buffer_insert_with_tags(ses_log_buf, &end, hex_data_buf, strlen(
		hex_data_buf), fg_tag, NULL);
	}

  }

void gui_add_log(char *buf, int color)
  {
  GtkTextBuffer *stat_buf;
  GtkTextIter end;
  GtkTextTag *fg_tag = NULL;
  GtkTextTagTable *tag_table;

  stat_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.status_text));
  gtk_text_buffer_get_end_iter(stat_buf, &end);
  tag_table = gtk_text_buffer_get_tag_table(stat_buf);

  /* choose a color */
  if (color == 0)
	gtk_text_buffer_insert(stat_buf, &end, buf, -1);
  else
	{
	if (color == 1)
	  fg_tag = gtk_text_tag_table_lookup(tag_table, "red_fg");
	else if (color == 2)
	  fg_tag = gtk_text_tag_table_lookup(tag_table, "red_fgs");
	else if (color == 3)
	  fg_tag = gtk_text_tag_table_lookup(tag_table, "blue_fgr");
	gtk_text_buffer_insert_with_tags(stat_buf, &end, buf, -1, fg_tag, NULL);
	}

  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(gui_data.status_text), &end, 0,
	  TRUE, 0, 0);

  gui_update();
  }

static void set_tab_name(char *text)
  {
  gint current_tab_num = 0;
  GtkWidget *current_lab, *current_tab;
  current_tab_num = gtk_notebook_get_current_page(GTK_NOTEBOOK(
	  gui_data.ses_tabs));
  current_tab = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gui_data.ses_tabs),
	  current_tab_num);
  current_lab = gtk_notebook_get_tab_label(GTK_NOTEBOOK(gui_data.ses_tabs),
	  current_tab);
  gtk_label_set_text(GTK_LABEL(current_lab), text);
  }

static const gchar *get_tab_name()
  {
  gint current_tab_num = 0;
  GtkWidget *current_lab, *current_tab;
  current_tab_num = gtk_notebook_get_current_page(GTK_NOTEBOOK(
	  gui_data.ses_tabs));
  current_tab = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gui_data.ses_tabs),
	  current_tab_num);
  current_lab = gtk_notebook_get_tab_label(GTK_NOTEBOOK(gui_data.ses_tabs),
	  current_tab);
  const gchar *text;
  text = gtk_label_get_text(GTK_LABEL(current_lab));
  return text;
  }

static void gui_file_open(struct tj_data *gui_inst_data)
  {

  /* open file dialog */
  GtkWidget *dialog;
  dialog = gtk_file_chooser_dialog_new("Open File", GTK_WINDOW(
	  gui_data.window), GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL,
	  GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog),
	  gui_data.cur_folder);

  if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
	{
	/* create a new tab */
	new_tab(gui_inst_data);

	char *file_name;
	file_name = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));

	/* load config file */
	strcpy(gui_inst_data->sess_file_name, file_name);
	load_session_data(gui_inst_data);

	/* disable unprintable chars */
	int i;
	int binary_found = 0;
	for (i = 0; i < gui_inst_data->sess_data_len; i++)
	  {
	  if (isprint(gui_inst_data->sess_data[i]) < 1
		  && isspace(gui_inst_data->sess_data[i]) < 1)
		{
		gui_inst_data->sess_data[i] = '.';
		binary_found = 1;
		}
	  }

	if (binary_found)
	  tj_error(
		  gui_inst_data,
		  "Warning: unprintable characters replaced with '.', only CLI mode supports binary files.",
		  "", c_red);

	/* load file in text box */
	GtkTextBuffer *buffer;
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
	gtk_text_buffer_set_text(buffer, gui_inst_data->sess_data,
		gui_inst_data->sess_data_len);

	/* display filename in tab */
	char *fn_only;
	for (fn_only = file_name + strlen(file_name); fn_only >= file_name
		&& *fn_only != '/'; fn_only--)
	  ;
	fn_only++;
	set_tab_name(fn_only);

	/* cleanup */
	g_free(file_name);
	free(gui_data.cur_folder);
	gui_data.cur_folder = gtk_file_chooser_get_current_folder(
		GTK_FILE_CHOOSER(dialog));
	}
  gtk_widget_destroy(dialog);

  }

static void gui_file_save_as(struct tj_data *gui_inst_data)
  {
  int i;
  FILE *save_file;
  GtkWidget *dialog;
  dialog = gtk_file_chooser_dialog_new("Save File", GTK_WINDOW(
	  gui_data.window), GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL,
	  GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog),
	  gui_data.cur_folder);
  gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(dialog), TRUE);

  if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
	{
	char *file_name;
	file_name = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
	GtkTextIter start, end;
	GtkTextBuffer *buffer;
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
	gtk_text_buffer_get_bounds(buffer, &start, &end);
	char *gui_session_text = gtk_text_buffer_get_text(buffer, &start, &end,
		FALSE);

	if ((save_file = fopen(file_name, "wb")) == NULL)
	  tj_error(gui_inst_data, "Can not save session file ", file_name, c_red);
	else
	  {
	  for (i = 0; i <= strlen(gui_session_text) - 1; i++)
		putc(gui_session_text[i], save_file);
	  fclose(save_file);
	  tj_error(gui_inst_data, "File saved as ", file_name, c_black);
	  }

	/* display filename in tab */
	char *fn_only;
	for (fn_only = file_name + strlen(file_name); fn_only >= file_name
		&& *fn_only != '/'; fn_only--)
	  ;
	fn_only++;
	set_tab_name(fn_only);

	/* cleanup */
	g_free(file_name);
	free(gui_data.cur_folder);
	gui_data.cur_folder = gtk_file_chooser_get_current_folder(
		GTK_FILE_CHOOSER(dialog));
	}
  gtk_widget_destroy(dialog);

  }

static void gui_file_save(struct tj_data *gui_inst_data)
  {
  int i;
  FILE *save_file;
  char *file_name = (char *) get_tab_name();
  char full_path[FILENAME_MAX];
  GtkTextIter start, end;
  sprintf(full_path, "%s/%s", gui_data.cur_folder, file_name);

  /* get text from buffer */
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  char *gui_session_text =
	  gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

  /* save file */
  if ((save_file = fopen(full_path, "wb")) == NULL)
	tj_error(gui_inst_data, "Can not save session file ", full_path, c_red);
  else
	{
	if (strlen(gui_session_text) > 0)
	  for (i = 0; i <= strlen(gui_session_text) - 1; i++)
		putc(gui_session_text[i], save_file);
	fclose(save_file);
	tj_error(gui_inst_data, "File saved as ", full_path, c_black);
	}

  }

static void ssl_check_toggled(struct tj_data *gui_inst_data,
	GtkToggleButton *ssl_checkbox)
  {
  if (gtk_toggle_button_get_active(ssl_checkbox))
	gui_inst_data->use_ssl = 1;
  else
	gui_inst_data->use_ssl = 0;
  }

static void wait_data_check_toggled(struct tj_data *gui_inst_data,
	GtkToggleButton *wait_data_checkbox)
  {
  if (gtk_toggle_button_get_active(wait_data_checkbox))
	gui_inst_data->wait_conn = 1;
  else
	gui_inst_data->wait_conn = 0;
  }

static void close_last_check_toggled(struct tj_data *gui_inst_data,
	GtkToggleButton *close_last_checkbox)
  {
  if (gtk_toggle_button_get_active(close_last_checkbox))
	gui_inst_data->no_wait_last = 1;
  else
	gui_inst_data->no_wait_last = 0;
  }

static void auto_scroll_toggled(GtkToggleButton *auto_scr_checkbox)
  {
  if (gtk_toggle_button_get_active(auto_scr_checkbox))
	gui_data.auto_scroll = 1;
  else
	gui_data.auto_scroll = 0;
  }

static void auto_clearstat_toggled(GtkToggleButton *auto_clearstat_checkbox)
  {
  if (gtk_toggle_button_get_active(auto_clearstat_checkbox))
	gui_data.auto_clearstat = 1;
  else
	gui_data.auto_clearstat = 0;
  }

static void hexdump_toggled(struct tj_data *gui_inst_data,
	GtkToggleButton *hexdump_checkbox)
  {
  if (gtk_toggle_button_get_active(hexdump_checkbox))
	gui_inst_data->hexdump = 1;
  else
	gui_inst_data->hexdump = 0;
  }

void gui_init(int argc, char *argv[], struct tj_data *gui_inst_data)
  {
  /* INIT */
  srand((unsigned int) (time(NULL)));
  gtk_init(&argc, &argv);

  gui_inst_data->gui = 1;
  gui_inst_data->verbose = 1;
  gui_inst_data->debug = 1;
  gui_data.auto_scroll = 0;
  gui_data.listening = 0;
  gui_data.auto_clearstat = 1;
  gui_data.cur_folder = getcwd(NULL, 0);
  gui_data.undo_buf = NULL;

  /* window */
  gui_data.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(gui_data.window), "Tcpjunk");
  gtk_container_set_border_width(GTK_CONTAINER(gui_data.window), 1);
  g_signal_connect(G_OBJECT(gui_data.window), "delete_event", G_CALLBACK(
		  exit_gui), NULL);
  gtk_window_set_icon_from_file(GTK_WINDOW(gui_data.window),
	  "/usr/share/pixmaps/tcpjunk.png", NULL);
  gtk_window_resize(GTK_WINDOW(gui_data.window), 640, 480);

  /* boxes */
  GtkWidget *main_box, *ctl_hbox, *right_box, *conf_box, *status_box;
  main_box = gtk_vbox_new(FALSE, 0);
  ctl_hbox = gtk_hbox_new(FALSE, 0);
  right_box = gtk_vbox_new(FALSE, 0);
  conf_box = gtk_vbox_new(FALSE, 0);
  status_box = gtk_vbox_new(FALSE, 0);

  /* check boxes */
  GtkWidget *ssl_checkbox, *wait_data_checkbox, *close_last_checkbox,
	  *auto_scr_checkbox, *auto_clearstat_checkbox, *hexdump_checkbox;
  ssl_checkbox = gtk_check_button_new_with_label("Use SSL");
  wait_data_checkbox
	  = gtk_check_button_new_with_label("Wait for data at first");
  close_last_checkbox = gtk_check_button_new_with_label(
	  "Don't wait for data at last");
  auto_scr_checkbox
	  = gtk_check_button_new_with_label("Auto scroll session log");
  auto_clearstat_checkbox
	  = gtk_check_button_new_with_label("Auto clear status");
  hexdump_checkbox = gtk_check_button_new_with_label("Hex dump");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(auto_clearstat_checkbox), TRUE);
  g_signal_connect_swapped(G_OBJECT(ssl_checkbox), "toggled", G_CALLBACK(
		  ssl_check_toggled), gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(wait_data_checkbox), "toggled",
	  G_CALLBACK(wait_data_check_toggled), gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(close_last_checkbox), "toggled",
	  G_CALLBACK(close_last_check_toggled), gui_inst_data);
  g_signal_connect(G_OBJECT(auto_scr_checkbox), "toggled", G_CALLBACK(
		  auto_scroll_toggled), NULL);
  g_signal_connect(G_OBJECT(auto_clearstat_checkbox), "toggled", G_CALLBACK(
		  auto_clearstat_toggled), NULL);
  g_signal_connect_swapped(G_OBJECT(hexdump_checkbox), "toggled", G_CALLBACK(
		  hexdump_toggled), gui_inst_data);

  /* tabs */
  gui_data.ses_tabs = gtk_notebook_new();

  /* textviews */
  gui_data.session_log_text = gtk_source_view_new();
  gui_data.status_text = gtk_text_view_new();
  gtk_text_view_set_editable(GTK_TEXT_VIEW(gui_data.session_log_text), FALSE);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(gui_data.status_text), FALSE);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(gui_data.session_log_text),
	  GTK_WRAP_CHAR);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(gui_data.status_text),
	  GTK_WRAP_WORD);
  g_signal_connect(GTK_TEXT_VIEW(gui_data.session_log_text),
	  "copy-clipboard", G_CALLBACK(gui_show_selection), gui_inst_data);

  /* test buffers tags */
  GtkTextBuffer *ses_log_buf, *ses_stat_buf;
  ses_log_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(
	  gui_data.session_log_text));
  ses_stat_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.status_text));
  gtk_text_buffer_create_tag(ses_log_buf, "red_fg", "foreground", "#7f0000",
	  "background", "#fbeded", NULL);
  gtk_text_buffer_create_tag(ses_log_buf, "blue_fg", "foreground", "#00007f",
	  "background", "#ededfb", NULL);
  gtk_text_buffer_create_tag(ses_log_buf, "yellow_hl", "background", "#afffaf",
	  NULL);
  gtk_text_buffer_create_tag(ses_stat_buf, "red_fg", "foreground", "#ff0000",
	  NULL);
  gtk_text_buffer_create_tag(ses_stat_buf, "red_fgs", "foreground", "#7f0000",
	  "background", "#fbeded", NULL);
  gtk_text_buffer_create_tag(ses_stat_buf, "blue_fgr", "foreground", "#00007f",
	  "background", "#ededfb", NULL);

  /* textviews fonts */
  PangoFontDescription *font_desc;
  font_desc = pango_font_description_from_string("Fixed 13");
  gtk_widget_modify_font(gui_data.session_log_text, font_desc);
  pango_font_description_free(font_desc);

  /* text entries */
  gui_data.host_entry = gtk_entry_new();
  gui_data.port_entry = gtk_entry_new();
  gui_data.sport_entry = gtk_entry_new();
  gui_data.init_to_entry = gtk_entry_new();
  gui_data.sock_to_entry = gtk_entry_new();
  gui_data.repeat_entry = gtk_entry_new();
  gui_data.syntax_entry = gtk_entry_new();
  char temp_set[128];
  sprintf(temp_set, "%d", gui_inst_data->conn_timeout);
  gtk_entry_set_text(GTK_ENTRY(gui_data.init_to_entry), temp_set);
  sprintf(temp_set, "%d", gui_inst_data->reply_timeout / 1000);
  gtk_entry_set_text(GTK_ENTRY(gui_data.sock_to_entry), temp_set);
  gtk_entry_set_text(GTK_ENTRY(gui_data.repeat_entry), "1");
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.init_to_entry), 2);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.sock_to_entry), 5);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.port_entry), 5);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.sport_entry), 5);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.host_entry), 16);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.repeat_entry), 4);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.syntax_entry), 14);
  g_signal_connect(G_OBJECT(gui_data.syntax_entry), "key-press-event",
	  G_CALLBACK(syntax_entry_key_press), gui_inst_data);

  /* labels */
  GtkWidget *host_lab, *port_lab, *sock_to_lab, *init_to_lab, *repeat_lab,
	  *syntax_lab, *sport_lab;
  host_lab = gtk_label_new("Hostname/IP");
  port_lab = gtk_label_new("Port");
  sport_lab = gtk_label_new("Source port");
  init_to_lab = gtk_label_new("Initial timeout (sec) ");
  sock_to_lab = gtk_label_new("Reply timeout (ms) ");
  repeat_lab = gtk_label_new("Repeat session");
  syntax_lab = gtk_label_new("Highlight");
  gtk_misc_set_alignment(GTK_MISC(host_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(port_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(sport_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(init_to_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(sock_to_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(repeat_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(syntax_lab), 0.0, 1.0);

  /* tables */
  GtkWidget *conf_table, *display_table, *actions_table;
  conf_table = gtk_table_new(12, 2, FALSE);
  display_table = gtk_table_new(4, 2, FALSE);
  actions_table = gtk_table_new(2, 1, FALSE);

  /* frames */
  GtkWidget *session_frame, *session_log_frame, *conf_frame, *status_frame,
	  *other_frame, *actions_frame;
  session_frame = gtk_frame_new(NULL);
  session_log_frame = gtk_frame_new(NULL);
  conf_frame = gtk_frame_new(" Connection ");
  status_frame = gtk_frame_new(NULL);
  other_frame = gtk_frame_new(" Options ");
  actions_frame = gtk_frame_new(" Action ");

  /* menus */
  GtkWidget *menu_bar;
  menu_bar = gtk_menu_bar_new();
  GtkWidget *file_menu, *insert_menu, *help_menu;
  GtkWidget *file_mi, *insert_mi, *help_mi;
  file_menu = gtk_menu_new();
  insert_menu = gtk_menu_new();
  help_menu = gtk_menu_new();
  file_mi = gtk_menu_item_new_with_label("File");
  insert_mi = gtk_menu_item_new_with_label("Insert");
  help_mi = gtk_menu_item_new_with_label("Help");

  /* file and help menu items */
  GtkWidget *new_mi, *save_mi, *saveas_mi, *quit_mi, *close_mi, *about_mi;
  new_mi = gtk_menu_item_new_with_label("New");
  gui_data.open_mi = gtk_menu_item_new_with_label("Open");
  save_mi = gtk_menu_item_new_with_label("Save");
  saveas_mi = gtk_menu_item_new_with_label("Save as");
  close_mi = gtk_menu_item_new_with_label("Close");
  quit_mi = gtk_menu_item_new_with_label("Quit");
  about_mi = gtk_menu_item_new_with_label("About");
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), new_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), close_mi);
  GtkWidget *sep1 = gtk_separator_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), sep1);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), gui_data.open_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), save_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), saveas_mi);
  GtkWidget *sep2 = gtk_separator_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), sep2);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), quit_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(help_menu), about_mi);

  /* insert menu items */
  GtkWidget *section_mi, *repeat_mi, *system_mi, *fuzz_mi, *enc_mi, *char_mi,
	  *count_mi;
  section_mi = gtk_menu_item_new_with_label("Section");
  repeat_mi = gtk_menu_item_new_with_label("Repeat");
  system_mi = gtk_menu_item_new_with_label("System");
  fuzz_mi = gtk_menu_item_new_with_label("Fuzz");
  enc_mi = gtk_menu_item_new_with_label("Encode");
  char_mi = gtk_menu_item_new_with_label("Charaters");
  count_mi = gtk_menu_item_new_with_label("Counter");
  gtk_menu_shell_append(GTK_MENU_SHELL(insert_menu), section_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(insert_menu), repeat_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(insert_menu), system_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(insert_menu), fuzz_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(insert_menu), char_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(insert_menu), enc_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(insert_menu), count_mi);

  /* menu bar */
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), file_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), insert_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), help_mi);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_mi), file_menu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(insert_mi), insert_menu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(help_mi), help_menu);

  /* menu signals */
  g_signal_connect_swapped(G_OBJECT(gui_data.open_mi), "activate",
	  G_CALLBACK(gui_file_open), gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(saveas_mi), "activate", G_CALLBACK(
		  gui_file_save_as), gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(save_mi), "activate", G_CALLBACK(
		  gui_file_save), gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(new_mi), "activate", G_CALLBACK(new_tab),
	  gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(close_mi), "activate", G_CALLBACK(
		  close_tab), gui_inst_data);
  g_signal_connect(G_OBJECT(quit_mi), "activate", G_CALLBACK(exit_gui), NULL);
  g_signal_connect_swapped(G_OBJECT(about_mi), "activate", G_CALLBACK(
		  gui_about), gui_inst_data);
  g_signal_connect(G_OBJECT(section_mi), "activate", G_CALLBACK(
		  gui_insert_sect), NULL);
  g_signal_connect(G_OBJECT(repeat_mi), "activate",
	  G_CALLBACK(gui_insert_rep), NULL);
  g_signal_connect(G_OBJECT(system_mi), "activate",
	  G_CALLBACK(gui_insert_sys), NULL);
  g_signal_connect(G_OBJECT(fuzz_mi), "activate",
	  G_CALLBACK(gui_insert_fuzz), NULL);
  g_signal_connect(G_OBJECT(char_mi), "activate",
	  G_CALLBACK(gui_insert_char), NULL);
  g_signal_connect(G_OBJECT(enc_mi), "activate", G_CALLBACK(gui_insert_enc),
	  NULL);
  g_signal_connect(G_OBJECT(count_mi), "activate", G_CALLBACK(
		  gui_insert_count), NULL);

  /* scroll windows */
  GtkWidget *ses_set_scr_win, *session_log_scr_win, *status_scr_win;
  ses_set_scr_win = gtk_scrolled_window_new(NULL, NULL);
  session_log_scr_win = gtk_scrolled_window_new(NULL, NULL);
  status_scr_win = gtk_scrolled_window_new(NULL, NULL);
  gtk_container_add(GTK_CONTAINER(session_log_scr_win),
	  gui_data.session_log_text);
  gtk_container_add(GTK_CONTAINER(status_scr_win), gui_data.status_text);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(ses_set_scr_win),
	  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(session_log_scr_win),
	  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(status_scr_win),
	  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

  /* buttons */
  gui_data.connect_butt = gtk_button_new_with_label("Connect [Ctrl+E]");
  gui_data.listen_butt = gtk_button_new_with_label("Listen [Ctrl+R]");
  g_signal_connect_swapped(G_OBJECT(gui_data.connect_butt), "clicked",
	  G_CALLBACK(gui_connect), gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(gui_data.listen_butt), "clicked",
	  G_CALLBACK(gui_listen), (gpointer) gui_inst_data);

  /* alignments */
  GtkWidget *port_alig, *init_to_alig, *sock_to_alig, *conf_table_alig,
	  *other_table_alig, *host_alig, *repeat_alig, *actions_table_alig,
	  *syntax_alig, *sport_alig, *conn_alig, *lis_alig;
  port_alig = gtk_alignment_new(0, 0, 0, 0);
  sport_alig = gtk_alignment_new(0, 0, 0, 0);
  host_alig = gtk_alignment_new(0, 0, 0, 0);
  init_to_alig = gtk_alignment_new(0, 0, 0, 0);
  sock_to_alig = gtk_alignment_new(0, 0, 0, 0);
  repeat_alig = gtk_alignment_new(0, 0, 0, 0);
  conf_table_alig = gtk_alignment_new(0, 0, 0, 0);
  other_table_alig = gtk_alignment_new(1, 1, 1, 0);
  actions_table_alig = gtk_alignment_new(0.5, 0, 1, 0);
  syntax_alig = gtk_alignment_new(0.5, 0, 0, 0);
  conn_alig = gtk_alignment_new(0.0, 0, 1, 0);
  lis_alig = gtk_alignment_new(0.0, 0, 1, 0);
  gtk_container_add(GTK_CONTAINER(port_alig), gui_data.port_entry);
  gtk_container_add(GTK_CONTAINER(sport_alig), gui_data.sport_entry);
  gtk_container_add(GTK_CONTAINER(host_alig), gui_data.host_entry);
  gtk_container_add(GTK_CONTAINER(init_to_alig), gui_data.init_to_entry);
  gtk_container_add(GTK_CONTAINER(sock_to_alig), gui_data.sock_to_entry);
  gtk_container_add(GTK_CONTAINER(conf_table_alig), conf_table);
  gtk_container_add(GTK_CONTAINER(actions_table_alig), actions_table);
  gtk_container_add(GTK_CONTAINER(other_table_alig), display_table);
  gtk_container_add(GTK_CONTAINER(repeat_alig), gui_data.repeat_entry);
  gtk_container_add(GTK_CONTAINER(syntax_alig), gui_data.syntax_entry);
  gtk_container_add(GTK_CONTAINER(conn_alig), gui_data.connect_butt);
  gtk_container_add(GTK_CONTAINER(lis_alig), gui_data.listen_butt);
  gtk_alignment_set_padding(GTK_ALIGNMENT(conf_table_alig), 3, 3, 5, 5);
  gtk_alignment_set_padding(GTK_ALIGNMENT(other_table_alig), 3, 3, 5, 5);
  gtk_alignment_set_padding(GTK_ALIGNMENT(actions_table_alig), 3, 3, 5, 5);

  /* vpanes */
  GtkWidget *sess_pan;
  sess_pan = gtk_vpaned_new();
  gtk_paned_set_position(GTK_PANED(sess_pan), 200);

  /* pack sessions texts */
  gtk_container_add(GTK_CONTAINER(session_frame), ses_set_scr_win);
  gtk_container_add(GTK_CONTAINER(session_log_frame), session_log_scr_win);
  gtk_container_add(GTK_CONTAINER(conf_frame), conf_table_alig);
  gtk_container_add(GTK_CONTAINER(status_frame), status_scr_win);
  gtk_box_pack_start(GTK_BOX(status_box), GTK_WIDGET(status_frame), TRUE, TRUE,
	  1);

  /* fill setup table */
  gtk_table_attach(GTK_TABLE(conf_table), host_lab, 0, 1, 0, 1, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), port_lab, 1, 2, 0, 1, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 0);
  gtk_table_attach(GTK_TABLE(conf_table), host_alig, 0, 1, 1, 2, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), port_alig, 1, 2, 1, 2, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), ssl_checkbox, 0, 2, 3, 4, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), wait_data_checkbox, 0, 2, 4, 5,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), close_last_checkbox, 0, 2, 5, 6,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), init_to_lab, 0, 1, 6, 7, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), init_to_alig, 1, 2, 6, 7, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), sock_to_lab, 0, 1, 7, 8, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), sock_to_alig, 1, 2, 7, 8, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), repeat_lab, 0, 1, 8, 9, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), repeat_alig, 1, 2, 8, 9, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), sport_lab, 0, 1, 9, 10, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 0);
  gtk_table_attach(GTK_TABLE(conf_table), sport_alig, 1, 2, 9, 10, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);

  /* pack display options table */
  gtk_table_attach(GTK_TABLE(display_table), hexdump_checkbox, 0, 2, 0, 1,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(display_table), auto_scr_checkbox, 0, 2, 1, 2,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(display_table), auto_clearstat_checkbox, 0, 2, 2,
	  3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(display_table), syntax_lab, 0, 1, 3, 4, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(display_table), syntax_alig, 1, 2, 3, 4,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_container_add(GTK_CONTAINER(other_frame), other_table_alig);

  /* pack actions table */
  gtk_table_attach(GTK_TABLE(actions_table), conn_alig, 0, 1, 0, 1, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(actions_table), lis_alig, 0, 1, 1, 2, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_container_add(GTK_CONTAINER(actions_frame), actions_table_alig);

  /* accelerators */
  GtkAccelGroup *ac_grp;
  ac_grp = gtk_accel_group_new();
  gtk_window_add_accel_group(GTK_WINDOW(gui_data.window), ac_grp);
  gtk_widget_add_accelerator(gui_data.connect_butt, "activate", ac_grp, GDK_E,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(gui_data.listen_butt, "activate", ac_grp, GDK_L,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(gui_data.open_mi, "activate", ac_grp, GDK_O,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(close_mi, "activate", ac_grp, GDK_W,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(new_mi, "activate", ac_grp, GDK_N,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(save_mi, "activate", ac_grp, GDK_S,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);

  /* pack conf box */
  gtk_box_pack_start(GTK_BOX(conf_box), GTK_WIDGET(conf_frame), FALSE, FALSE, 1);
  gtk_box_pack_start(GTK_BOX(conf_box), GTK_WIDGET(other_frame), FALSE, FALSE,
	  1);
  gtk_box_pack_start(GTK_BOX(conf_box), GTK_WIDGET(actions_frame), FALSE,
	  FALSE, 1);

  /* pack right box */
  gtk_paned_add1(GTK_PANED(sess_pan), gui_data.ses_tabs);
  gtk_paned_add2(GTK_PANED(sess_pan), session_log_frame);
  gtk_box_pack_start(GTK_BOX(right_box), GTK_WIDGET(sess_pan), TRUE, TRUE, 1);
  gtk_box_pack_start(GTK_BOX(right_box), GTK_WIDGET(status_box), FALSE, FALSE,
	  1);

  /* pack window */
  gtk_box_pack_start(GTK_BOX(ctl_hbox), GTK_WIDGET(conf_box), FALSE, FALSE, 1);
  gtk_box_pack_start(GTK_BOX(ctl_hbox), GTK_WIDGET(right_box), TRUE, TRUE, 1);
  gtk_box_pack_start(GTK_BOX(main_box), GTK_WIDGET(menu_bar), FALSE, FALSE, 1);
  gtk_box_pack_start(GTK_BOX(main_box), GTK_WIDGET(ctl_hbox), TRUE, TRUE, 1);
  gtk_container_add(GTK_CONTAINER(gui_data.window), main_box);
  new_tab(gui_inst_data);
  gtk_widget_show_all(gui_data.window);

  /* show version */
  tj_error(gui_inst_data, "Tcpjunk ", VER, c_black);

  gtk_main();

  }

