/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include "tcpjunk.h"

void junk(struct tj_data *inst_data)
  {

  /* setup session send buffer */
  int send_buf_len = inst_data->sess_data_len;
  char *send_buf = NULL;
  send_buf = tj_malloc(send_buf, send_buf_len);

  /* other stuff */
  int buf_id;
  int loop_counter = inst_data->loop;
  int keepalive_counter = inst_data->keepalive;
  int rc;
  int cl_socket;
  int r_timeout;

  /* socket stuff */
  struct sockaddr_in sa;
  memset(&sa, 0, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_port = htons(inst_data->dest_port);
  memcpy(&sa.sin_addr.s_addr, inst_data->dest_addr, inst_data->dest_addr_len);

  /* SSL Stuff */
  int err;
  SSL_CTX *ctx = NULL;
  SSL *ssl = NULL;
  SSL_METHOD *meth;

  /* Prepare SSL Stuff if needed */
  if (inst_data->use_ssl == 1)
    {
    SSL_library_init();
    SSL_load_error_strings();

    /* Create SSL_CTX structure */
    meth = SSLv3_method();
    ctx = SSL_CTX_new(meth);
    if ((ctx) == NULL)
      {
      tj_error(inst_data, "Cannot create an SSL_CTX structure", "", c_red);
      exit(EXIT_FAILURE);
      }
    /* SSL structure */
    ssl = SSL_new(ctx);
    if ((ssl) == NULL)
      {
      tj_error(inst_data, "Cannot create an SSL structure", "", c_red);
      exit(EXIT_FAILURE);
      }
    }

  /* main loop */
  while (inst_data->loop == 0 ? 1 : loop_counter--)
    {

    /* load session data file */
    if (inst_data->auto_file > 0)
      {
      /* choose random file */
      set_session_file(inst_data);

      /* load session file */
      load_session_data(inst_data);

      /* check config */
      find_sections(inst_data);
      }

    /* allocate a socket */
    cl_socket = socket(AF_INET, SOCK_STREAM, 0);
    if (cl_socket < 0)
      tj_error(inst_data, "Socket allocation failed: ", strerror(errno), c_red);

    /* bind to different source IP address */
    if (strcmp(inst_data->interface, "") != 0)
      {
      struct sockaddr_in ba;
      memset(&ba, 0, sizeof(ba));
      ba.sin_family = AF_INET;
      char bind_if_name[32];
      char new_src_ip[32];

      /* setup alias */
      sprintf(bind_if_name, "%s:%d", inst_data->interface, inst_data->inst_num
          + inst_data->offset - 1);

      if (inst_data->rand_ip == 1)
        /* set random IP */
        sprintf(new_src_ip, "%d.%d.%d.%d", rand() % 224, rand() % 224, rand()
            % 224, rand() % 224);
      else
        /* setup specific IP */
        sprintf(new_src_ip, inst_data->src_ip);

      /* make sure address is cool */
      check_src_address(inst_data, new_src_ip);

      /* setup IP on alias */
      set_src_ip(inst_data, bind_if_name, new_src_ip);

      /* setup source port if needed */
      if (inst_data->src_port > -1)
        ba.sin_port = ntohs(inst_data->src_port);

      /* bind socket to IP address */
      memcpy(&ba.sin_addr.s_addr, inst_data->src_addr, inst_data->src_addr_len);
      if (bind(cl_socket, (struct sockaddr *) &ba, sizeof(struct sockaddr_in))
          < 0)
        tj_error(inst_data, "Socket binding failed: ", strerror(errno), c_red);
      }
    else if (inst_data->src_port > -1)
      {
      /* only bind to different source port */
      struct sockaddr_in ba;
      memset(&ba, 0, sizeof(ba));
      ba.sin_family = AF_INET;
      sa.sin_addr.s_addr = INADDR_ANY;
      ba.sin_port = ntohs(inst_data->src_port);

      /* bind socket to IP address */
      memcpy(&ba.sin_addr.s_addr, inst_data->src_addr, inst_data->src_addr_len);
      if (bind(cl_socket, (struct sockaddr *) &ba, sizeof(struct sockaddr_in))
          < 0)
        tj_error(inst_data, "Socket binding failed: ", strerror(errno), c_red);
      }

    /* connect */
    rc = tj_connect(cl_socket, sa, inst_data);

    /* if no error, start transfers */
    if (rc == 0)
      {
      if (inst_data->use_ssl == 1)
        {
        /* Assign socket to SSL structure */
        SSL_set_fd(ssl, cl_socket);

        /* SSL Handshake */
        err = SSL_connect(ssl);
        if ((err) == -1)
          tj_error(inst_data, "SSL connect ", ERR_error_string(ERR_get_error(),
              NULL), c_red);
        }
      send_data:

      buf_id = 0;
      r_timeout = 0;

      /* get next send buffer from config */
      while (buf_id < inst_data->sect_idx)
        {
        /* see what client has to say on start */
        if (buf_id == 0 && inst_data->wait_conn == 1)
          r_timeout = read_reply(cl_socket, ssl, inst_data);
        if (r_timeout > 0)
          {
          tj_error(inst_data, "Reply timeout", "", c_red);
          goto close_con;
          }

        /* prepare a temp send buffer */
        if (send_buf_len != inst_data->sects_len[buf_id])
          {
          send_buf_len = inst_data->sects_len[buf_id];
          send_buf = tj_realloc(send_buf, send_buf_len);
          }

        memset(send_buf, 0, send_buf_len);
        memcpy(send_buf, inst_data->sects[buf_id], send_buf_len);

        /* process tags */
        send_buf_len = repeat_tag(&send_buf, inst_data, send_buf_len);
        send_buf_len = instance_tag(&send_buf, inst_data, send_buf_len);
        send_buf_len = count_tag(&send_buf, inst_data, send_buf_len);
        send_buf_len = char_tag(&send_buf, inst_data, send_buf_len);
        send_buf_len = fuzz_tag(&send_buf, inst_data, send_buf_len);
        send_buf_len = system_tag(&send_buf, inst_data, send_buf_len);
        send_buf_len = encode_tag(&send_buf, inst_data, send_buf_len);

        /* send data*/
        send_data(cl_socket, ssl, inst_data, send_buf, send_buf_len);

        /* see what server has to say, if we need to */
        if (!(inst_data->no_wait_last == 1 && buf_id == inst_data->sect_idx - 1))
          r_timeout = read_reply(cl_socket, ssl, inst_data);
        if (r_timeout > 0)
          {
          tj_error(inst_data, "Reply timeout", "", c_red);
          goto close_con;
          }

        buf_id++;
        }

      /* if keep alive send again */
      if (keepalive_counter > 1 && r_timeout == 0)
        {
        keepalive_counter--;
        goto send_data;
        }

      /* close socket */
      close_con:

      tj_disconnect(cl_socket, inst_data, ssl, ctx);

      if (inst_data->debug == 1)
        goto exit_child;
      }
    else
      {
      /* connection failed */
      tj_error(inst_data, "Connection failed: ", strerror(errno), c_red);
      tj_disconnect(cl_socket, inst_data, ssl, ctx);
      }

    /* wait if need to */
    if (inst_data->wait > 0)
      usleep(inst_data->wait);

    /* stop if only one instance debug mode */
    if (inst_data->debug == 1)
      break;

    }
  /* end of main loop */
  exit_child:

  if (inst_data->use_ssl == 1)
    {
    err = SSL_shutdown(ssl);
    if ((err) == -1)
      tj_error(inst_data, "SSL Disconnect ", ERR_error_string(ERR_get_error(),
          NULL), c_red);
    SSL_free(ssl);
    SSL_CTX_free(ctx);
    }

  free(send_buf);
  if (inst_data->gui == 0)
    exit(EXIT_SUCCESS);
  else
    return;

  }

