/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include "tcpjunk.h"

void sjunk(struct tj_data *inst_data)
  {
  /* some vars */
  int r_timeout;
  int buf_id;
  int send_buf_len = inst_data->sess_data_len;
  char *send_buf = NULL;
  send_buf = tj_malloc(send_buf, send_buf_len);
  int keepalive_counter = inst_data->keepalive;

  /* SSL Stuff */
  int err;
  SSL_CTX *ctx = NULL;
  SSL *ssl = NULL;
  SSL_METHOD *meth;

  /* prepare SSL stuff if needed */
  if (inst_data->use_ssl == 1)
	{
	SSL_library_init();
	SSL_load_error_strings();

	/* SSL_METHOD structure */
	meth = SSLv3_method();
	ctx = SSL_CTX_new(meth);
	if ((ctx) == NULL)
	  {
	  printf("Cannot create an SSL_CTX structure\n");
	  goto finish;
	  }

	/* load server certificate */
	if (SSL_CTX_use_certificate_file(ctx, inst_data->server_cert,
		SSL_FILETYPE_PEM) <= 0)
	  {
	  tj_error(inst_data, "SSL certificate ", ERR_error_string(ERR_get_error(),
		  NULL), c_red);
	  goto finish;
	  }

	/* load private-key */
	if (SSL_CTX_use_PrivateKey_file(ctx, inst_data->server_key,
		SSL_FILETYPE_PEM) <= 0)
	  {
	  tj_error(inst_data, "SSL key ", ERR_error_string(ERR_get_error(), NULL),
		  c_red);
	  goto finish;
	  }

	/* check certificate and key match */
	if (!SSL_CTX_check_private_key(ctx))
	  {
	  tj_error(inst_data, "SSL - ", "Key does not match certificate", c_red);
	  goto finish;
	  }

	/* SSL structure */
	ssl = SSL_new(ctx);

	if ((ssl) == NULL)
	  {
	  tj_error(inst_data, "Cannot create SSL structure", "", c_red);
	  goto finish;
	  }

	/* Assign socket to SSL structure */
	SSL_set_fd(ssl, inst_data->srv_sock);

	/* SSL Handshake */
	err = SSL_accept(ssl);
	if ((err) == -1)
	  {
	  tj_error(inst_data, "SSL ", ERR_error_string(ERR_get_error(), NULL),
		  c_red);
	  goto finish;
	  }
	}

  /* start transfer */
  send_data:

  buf_id = 0;
  r_timeout = 0;

  /* get next send buffer from file */
  while (buf_id < inst_data->sect_idx)
	{
	/* see what client has to say on start */
	if (buf_id == 0 && inst_data->wait_conn == 1)
	  r_timeout = read_reply(inst_data->srv_sock, ssl, inst_data);
	if (r_timeout > 0)
	  {
	  tj_error(inst_data, "No data received, closing connection", "", c_red);
	  goto finish;
	  }

	/* prepare a temp send buffer */
	if (send_buf_len != inst_data->sects_len[buf_id])
	  {
	  send_buf_len = inst_data->sects_len[buf_id];
	  send_buf = tj_realloc(send_buf, send_buf_len);
	  }

	memcpy(send_buf, inst_data->sects[buf_id], send_buf_len);

	/* process tags */
	send_buf_len = repeat_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = instance_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = count_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = char_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = fuzz_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = system_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = encode_tag(&send_buf, inst_data, send_buf_len);

	/* send data to client */
	send_data(inst_data->srv_sock, ssl, inst_data, send_buf, send_buf_len);

	/* see what client has to say, if we need to */
	if (!(inst_data->no_wait_last == 1 && buf_id == inst_data->sect_idx - 1))
	  {
	  r_timeout = read_reply(inst_data->srv_sock, ssl, inst_data);
	  if (r_timeout > 0)
		{
		tj_error(inst_data, "No data received, closing connection", "", c_red);
		goto finish;
		}
	  }

	buf_id++;
	}

  /* if keep alive send again */
  if (keepalive_counter > 1 && r_timeout == 0)
	{
	keepalive_counter--;
	goto send_data;
	}

  /* complete */
  finish:

  free(send_buf);
  tj_disconnect(inst_data->srv_sock, inst_data, ssl, ctx);
  if (inst_data->gui == 0)
	exit(EXIT_SUCCESS);
  else
	return;

  }

