/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include "tcpjunk.h"

enum fuzz_types
  {
  r_none,
  r_digit,
  r_alnum,
  r_alpha,
  r_cntrl,
  r_print,
  r_special,
  r_punct,
  r_space,
  r_any,
  r_random
  };

enum dig_types
  {
  e_none,
  e_dec,
  e_hex,
  e_HEX,
  e_oct
  };

/* encoding */
void encode(char *dest, char *src, int src_size, char *enc_pre, int enc_type,
	char *enc_post, int enc_pad)
  {
  int dest_loc = 0;
  int src_loc;
  char tmp_char[16] = "", tmp_spr[16] = "";

  /* determine prefix and suffix */
  int prefix = 1;
  int postfix = 1;
  if (strcmp(enc_pre, "none") == 0)
	prefix = 0;
  if (strcmp(enc_post, "none") == 0)
	postfix = 0;

  /* encoding loop */
  for (src_loc = 0; src_loc < src_size; src_loc++)
	{
	  if (prefix)
		{
		memcpy(&dest[dest_loc], enc_pre, strlen(enc_pre));
		dest_loc += strlen(enc_pre);
		}

	  switch (enc_type)
		{
	  case e_dec:
		sprintf(tmp_spr, "%%0%dd", enc_pad);
		break;
	  case e_hex:
		sprintf(tmp_spr, "%%0%dx", enc_pad);
		break;
	  case e_HEX:
		sprintf(tmp_spr, "%%0%dX", enc_pad);
		break;
	  case e_oct:
		sprintf(tmp_spr, "%%0%do", enc_pad);
		break;
		}
	  sprintf(tmp_char, tmp_spr, (unsigned char) src[src_loc]);

	  memcpy(&dest[dest_loc], tmp_char, strlen(tmp_char));
	  dest_loc += strlen(tmp_char);

	  if (postfix)
		{
		memcpy(&dest[dest_loc], enc_post, strlen(enc_post));
		dest_loc += strlen(enc_post);
		}
	}
  dest[dest_loc] = '\0';
  }

/* check fuzz type */
int get_fuzz_type(char *fuzz_type, struct tj_data *inst_data)
  {
  if (strcmp(fuzz_type, "none") == 0)
	return r_none;
  else if (strcmp(fuzz_type, "digit") == 0)
	return r_digit;
  else if (strcmp(fuzz_type, "alnum") == 0)
	return r_alnum;
  else if (strcmp(fuzz_type, "alpha") == 0)
	return r_alpha;
  else if (strcmp(fuzz_type, "cntrl") == 0)
	return r_cntrl;
  else if (strcmp(fuzz_type, "print") == 0)
	return r_print;
  else if (strcmp(fuzz_type, "special") == 0)
	return r_special;
  else if (strcmp(fuzz_type, "punct") == 0)
	return r_punct;
  else if (strcmp(fuzz_type, "space") == 0)
	return r_space;
  else if (strcmp(fuzz_type, "any") == 0)
	return r_any;
  else if (strcmp(fuzz_type, "random") == 0)
	return r_random;
  else
	{
	tj_error(inst_data, "Invalid fuzz type - ", fuzz_type, c_red);
	if (inst_data->gui == 0)
	  exit(EXIT_FAILURE);
	else
	  return 0;
	}
  }

char fuzz_char(int ftype)
  {

  int rg = 0;
  int tr = 0;

  /* digit 48-57 */
  if (ftype == r_digit)
	return rand() % 10 + 48;

  /* alnum 48-57 65-90 97-122 */
  else if (ftype == r_alnum)
	{
	rg = rand() % 3;
	if (rg == 0)
	  return rand() % 10 + 48;
	else if (rg == 1)
	  return rand() % 26 + 65;
	else if (rg == 2)
	  return rand() % 26 + 97;
	}

  /* alpha 65-90 97-122 */
  else if (ftype == r_alpha)
	{
	rg = rand() % 2;
	if (rg == 0)
	  return rand() % 26 + 65;
	else if (rg == 1)
	  return rand() % 26 + 97;
	}

  /* cntrl 1-31 127 */
  if (ftype == r_cntrl)
	{
	tr = rand() % 33 + 1;
	if (tr == 32)
	  tr = 127;
	return tr;
	}

  /* print 32-126 */
  if (ftype == r_print)
	return rand() % 95 + 32;

  /* special 128-255 */
  if (ftype == r_special)
	return rand() % 128 + 128;

  /* punct 33-47 58-64 91-96 123-126 */
  else if (ftype == r_punct)
	{
	rg = rand() % 4;
	if (rg == 0)
	  return rand() % 15 + 33;
	else if (rg == 1)
	  return rand() % 7 + 58;
	else if (rg == 2)
	  return rand() % 6 + 91;
	else if (rg == 3)
	  return rand() % 4 + 123;
	}

  /* space 9-13 32 */
  if (ftype == r_space)
	{
	tr = rand() % 6 + 9;
	if (tr == 14)
	  tr = 32;
	return tr;
	}

  /* any 0-255 */
  if (ftype == r_any)
	return rand() % 255;

  /* we never reach here */
  return 0;
  }

/* check digit type */
int get_dig_type(char *enc_type, struct tj_data *inst_data)
  {
  if (strcmp(enc_type, "dec") == 0)
	return e_dec;
  else if (strcmp(enc_type, "hex") == 0)
	return e_hex;
  else if (strcmp(enc_type, "HEX") == 0)
	return e_HEX;
  else if (strcmp(enc_type, "oct") == 0)
	return e_oct;
  else
	{
	tj_error(inst_data, "Invalid encoding type - ", enc_type, c_red);
	if (inst_data->gui == 0)
	  exit(EXIT_FAILURE);
	else
	  return 0;
	}
  }

int system_tag(char **send_buf_p, struct tj_data *instance_data,
	int send_buf_len)
  {
  char *send_buf = *send_buf_p;
  char *send_buf_loc = send_buf;
  char *system_tag = "<sys>";
  char *system_tag_end = "</sys>";
  char *system_tag_loc = NULL;

  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc = prep_buf;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;

  char system_cmd[FILENAME_MAX];
  char *system_cmd_loc;
  int tags_processed = 0;
  FILE *cmd_pipe;
  int pipe_char;
  int cmd_out_loc, cmd_out_size = 1024;
  char *cmd_out = NULL;
  cmd_out = tj_malloc(cmd_out, cmd_out_size);

  while ((system_tag_loc = memmem(send_buf_loc, (send_buf + send_buf_len
	  - send_buf_loc), system_tag, strlen(system_tag))))
	{
	/* add text up to tag */
	memcpy(prep_buf_loc, send_buf_loc, system_tag_loc - send_buf_loc);

	/* advance prep_buf_loc to end */
	prep_buf_loc += system_tag_loc - send_buf_loc;

	/* advance send_buf_loc to end of tag */
	send_buf_loc = system_tag_loc + strlen(system_tag);

	/* point to start of tag data */
	system_cmd_loc = send_buf_loc;

	/* advance send_buf_loc to end of tag data */
	if (!(send_buf_loc = memmem(send_buf_loc, (send_buf + send_buf_len
		- send_buf_loc), system_tag_end, strlen(system_tag_end))))
	  goto done;

	tags_processed = 1;

	/* parse tag */
	memset(system_cmd, 0, FILENAME_MAX);
	strncpy(system_cmd, system_cmd_loc, send_buf_loc - system_cmd_loc);

	/* advance send_buf_loc to end of closing tag */
	send_buf_loc += strlen(system_tag_end);

	/* get stuff from the pipe */
	cmd_out_loc = 0;
	if ((cmd_pipe = popen(system_cmd, "r")) != NULL)
	  {
	  while ((pipe_char = getc(cmd_pipe)) != EOF)
		{
		cmd_out[cmd_out_loc++] = pipe_char;
		if (cmd_out_loc == cmd_out_size)
		  {
		  cmd_out_size *= 2;
		  cmd_out = tj_realloc(cmd_out, cmd_out_size);
		  }
		}
	  /* resize prep_buf and add data */
	  prep_buf_loc_offset = prep_buf_loc - prep_buf;
	  prep_buf_len += (cmd_out_loc - strlen(system_tag)
		  - strlen(system_tag_end) - strlen(system_cmd));
	  prep_buf = tj_realloc(prep_buf, prep_buf_len);
	  prep_buf_loc = prep_buf + prep_buf_loc_offset;
	  memcpy(prep_buf_loc, cmd_out, cmd_out_loc);
	  prep_buf_loc += cmd_out_loc;

	  (void) pclose(cmd_pipe);
	  }
	}

  if (tags_processed)
	{
	/* add last part */
	memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

	/* copy prep_buf to send_buf */
	send_buf_len = prep_buf_len;
	send_buf = tj_realloc(send_buf, send_buf_len);
	memcpy(send_buf, prep_buf, send_buf_len);
	*send_buf_p = send_buf;
	}

  done:

  free(cmd_out);
  free(prep_buf);
  return send_buf_len;
  }

int repeat_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len)
  {
  char *send_buf = *send_buf_p;
  char *send_buf_loc;
  char *rep_tag = "<rep ";
  char *rep_tag_end = "</rep>";
  char *rep_tag_loc = NULL;
  char *rep_text_loc = NULL;
  int rep_text_len = 0;
  char tag_data[128];
  int rep_min;
  int rep_max;
  int repeats;
  int tags_processed;
  int vals, nesting;
  char *rep_tag_nest_loc = NULL;

  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;

  start:

  nesting = 0;
  tags_processed = 0;
  prep_buf_loc = prep_buf;
  send_buf_loc = send_buf;

  while ((rep_tag_loc = memmem(send_buf_loc, (send_buf + send_buf_len
	  - send_buf_loc), rep_tag, strlen(rep_tag))))
	{
	tags_processed = 1;

	/* check for nesting tags */
	while ((rep_tag_nest_loc = memmem(rep_tag_loc + strlen(rep_tag), (send_buf
		+ send_buf_len - (rep_tag_loc + strlen(rep_tag))), rep_tag, strlen(
		rep_tag))))

	  {
	  rep_tag_loc = rep_tag_nest_loc;
	  nesting = 1;
	  }

	/* add text up to tag */
	memcpy(prep_buf_loc, send_buf_loc, rep_tag_loc - send_buf_loc);

	/* advance prep_buf_loc to end */
	prep_buf_loc += (rep_tag_loc - send_buf_loc);

	/* advance send_buf_loc to end of tag */
	send_buf_loc = rep_tag_loc + strlen(rep_tag);

	/* location of repeat text */
	if (!(rep_text_loc = memmem(send_buf_loc, send_buf + send_buf_len
		- send_buf_loc, ">", 1)))
	  goto end;
	rep_text_loc++;

	/* store tag data */
	memset(&tag_data, 0, 128);
	strncpy(tag_data, rep_tag_loc + strlen(rep_tag), rep_text_loc - rep_tag_loc
		- strlen(rep_tag));

	/* parse repeat tag data */
	vals = sscanf(tag_data, "%d %d", &rep_min, &rep_max);
	if (vals < 1)
	  goto end;
	if (vals == 1 && rep_min < 1)
	  goto end;
	if (vals == 2 && (rep_min < 1 || rep_max < 1 || rep_min > rep_max))
	  goto end;

	/* advance send_buf_loc to end of tag data */
	if (!(send_buf_loc = memmem(send_buf_loc, (send_buf + send_buf_len
		- send_buf_loc), rep_tag_end, strlen(rep_tag_end))))
	  goto end;

	/* repeat text must exist before closing tag */
	if (rep_text_loc > send_buf_loc)
	  goto end;

	/* calculate repeat text length */
	rep_text_len = send_buf_loc - rep_text_loc;
	if (rep_text_len == 0)
	  goto end;

	/* set number of repetitions */
	if (vals == 2)
	  repeats = rand() % ((1 + rep_max - rep_min)) + rep_min;
	else
	  repeats = rep_min;

	/* advance send_buf_loc to end of end_tag */
	send_buf_loc += strlen(rep_tag_end);

	/* resize prep_buf */
	prep_buf_loc_offset = prep_buf_loc - prep_buf;
	prep_buf_len = prep_buf_len + repeats * rep_text_len - (send_buf_loc
		- rep_tag_loc);
	prep_buf = tj_realloc(prep_buf, prep_buf_len);
	prep_buf_loc = prep_buf + prep_buf_loc_offset;

	/* start repetition */
	int i;
	for (i = 1; i <= repeats; i++)
	  {
	  memcpy(prep_buf_loc, rep_text_loc, rep_text_len);
	  prep_buf_loc += rep_text_len;
	  }
	}

  if (tags_processed)
	{

	/* add last part from send buf to prep buf */
	memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

	/* copy prep buf to send buf */
	send_buf_len = prep_buf_len;
	send_buf = tj_realloc(send_buf, send_buf_len);
	memcpy(send_buf, prep_buf, send_buf_len);
	}

  if (nesting == 1)
	goto start;
  else
	end:

	free(prep_buf);
  *send_buf_p = send_buf;
  return send_buf_len;
  }

int instance_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len)
  {
  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc = prep_buf;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;

  char *send_buf = *send_buf_p;
  char *send_buf_loc = send_buf;
  char *inst_tag = "<inst>";
  char *inst_tag_loc = NULL;
  int tags_processed = 0;
  char instance_str[8];

  while ((inst_tag_loc = memmem(send_buf_loc, (send_buf + send_buf_len
	  - send_buf_loc), inst_tag, strlen(inst_tag))))
	{
	tags_processed = 1;

	/* add text up to tag */
	memcpy(prep_buf_loc, send_buf_loc, inst_tag_loc - send_buf_loc);

	/* advance prep_buf_loc to end */
	prep_buf_loc += (inst_tag_loc - send_buf_loc);

	/* advance send_buf_loc to end of tag */
	send_buf_loc = inst_tag_loc + strlen(inst_tag);

	/* insert instance number */
	sprintf(instance_str, "%d", inst_data->inst_num);
	memcpy(prep_buf_loc, instance_str, strlen(instance_str));
	prep_buf_loc += strlen(instance_str);

	/* resize prep_buf */
	prep_buf_loc_offset = prep_buf_loc - prep_buf;
	prep_buf_len = prep_buf_len - (send_buf_loc - inst_tag_loc) + 1;
	prep_buf = tj_realloc(prep_buf, prep_buf_len);
	prep_buf_loc = prep_buf + prep_buf_loc_offset;
	}

  if (tags_processed)
	{
	/* add last part */
	memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

	/* copy prep buf to send buf */
	send_buf_len = prep_buf_len;
	send_buf = tj_realloc(send_buf, send_buf_len);
	memcpy(send_buf, prep_buf, send_buf_len);
	}
  free(prep_buf);
  *send_buf_p = send_buf;
  return send_buf_len;
  }

int char_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len)
  {
  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc = prep_buf;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;

  char *send_buf = *send_buf_p;
  char *send_buf_loc = send_buf;
  char *char_tag = "<char ";
  char *char_tag_loc = NULL;
  char tag_data[128];
  int i, tags_processed = 0;
  int vals, char_hex, count;

  while ((char_tag_loc = memmem(send_buf_loc, (send_buf + send_buf_len
	  - send_buf_loc), char_tag, strlen(char_tag))))
	{
	tags_processed = 1;

	/* add text up to tag */
	memcpy(prep_buf_loc, send_buf_loc, char_tag_loc - send_buf_loc);

	/* advance prep_buf_loc to end */
	prep_buf_loc += (char_tag_loc - send_buf_loc);

	/* advance send_buf_loc to end of tag */
	send_buf_loc = char_tag_loc + strlen(char_tag);

	/* advance send_buf_loc to end of tag */
	if (!(send_buf_loc = memmem(send_buf_loc, send_buf + send_buf_len
		- send_buf_loc, ">", 1)))
	  goto done;
	send_buf_loc++;

	/* store tag data */
	memset(&tag_data, 0, 128);
	strncpy(tag_data, char_tag_loc + strlen(char_tag), send_buf_loc
		- char_tag_loc - strlen(char_tag));

	/* get tag data */
	vals = sscanf(tag_data, "%x %d", &char_hex, &count);
	if (vals == 1 || count < 1)
	  count = 1;

	/* resize prep_buf */
	prep_buf_loc_offset = prep_buf_loc - prep_buf;
	prep_buf_len = prep_buf_len + count - (send_buf_loc - char_tag_loc);
	prep_buf = tj_realloc(prep_buf, prep_buf_len);
	prep_buf_loc = prep_buf + prep_buf_loc_offset;

	/* insert characters */
	for (i = 1; i <= count; i++)
	  {
	  *prep_buf_loc = (unsigned char) char_hex;
	  prep_buf_loc++;
	  }
	}

  if (tags_processed)
	{
	/* add last part */
	memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

	/* copy prep buf to send buf */
	send_buf_len = prep_buf_len;
	send_buf = tj_realloc(send_buf, send_buf_len);
	memcpy(send_buf, prep_buf, send_buf_len);
	}

  *send_buf_p = send_buf;
  done:

  free(prep_buf);
  return send_buf_len;
  }

int fuzz_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len)
  {
  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc = prep_buf;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;

  char *send_buf = *send_buf_p;
  char *send_buf_loc = send_buf;
  char *fuzz_tag = "<fuzz ";
  char *fuzz_tag_loc = NULL;
  char tag_data[128];
  char tok_fuzz_type[16];
  int fuzz_buf_len = 0;
  int tags_processed = 0;
  int i, vals;
  int fuzz_tag_min, fuzz_tag_max;
  int fuzz_type = 0;
  char *last_fuzz_loc = NULL;
  int last_fuzz_offset = 0;
  int do_last;

  while ((fuzz_tag_loc = memmem(send_buf_loc, (send_buf + send_buf_len
	  - send_buf_loc), fuzz_tag, strlen(fuzz_tag))))
	{
	tags_processed = 1;
	do_last = 0;

	/* add text up to tag */
	memcpy(prep_buf_loc, send_buf_loc, fuzz_tag_loc - send_buf_loc);

	/* advance prep_buf_loc to end */
	prep_buf_loc += (fuzz_tag_loc - send_buf_loc);

	/* advance send_buf_loc to end of tag */
	send_buf_loc = fuzz_tag_loc + strlen(fuzz_tag);

	/* advance send_buf_loc to end of tag */
	if (!(send_buf_loc = memmem(send_buf_loc, send_buf + send_buf_len
		- send_buf_loc, ">", 1)))
	  goto done;
	send_buf_loc++;

	/* store tag data */
	memset(&tag_data, 0, 128);
	strncpy(tag_data, fuzz_tag_loc + strlen(fuzz_tag), send_buf_loc
		- fuzz_tag_loc - strlen(fuzz_tag));

	/* parse tag data */
	vals = sscanf(tag_data, "%s %d %d", tok_fuzz_type, &fuzz_tag_min,
		&fuzz_tag_max);

	/* set to just repeat last tag */
	if (strcmp(tok_fuzz_type, "last>") == 0)
	  do_last = 1;

	/* figure out fuzzing type */
	if (!do_last)
	  fuzz_type = get_fuzz_type(tok_fuzz_type, inst_data);

	/* quick tag syntax check */
	if ((fuzz_type == 0 && !do_last) || (!do_last && (fuzz_tag_min < 1
		|| (fuzz_tag_min > fuzz_tag_max && vals == 3))))
	  {
	  tj_error(inst_data, "syntax error in <fuzz> tag ", "", c_red);
	  goto done;
	  }

	/* set fuzzing buffer length */
	if (!do_last)
	  {
	  if (vals == 3)
		fuzz_buf_len = rand() % (1 + (fuzz_tag_max - fuzz_tag_min))
			+ fuzz_tag_min;
	  else
		fuzz_buf_len = fuzz_tag_min;
	  }

	/* save last_fuzz_loc */
	if (last_fuzz_loc)
	  last_fuzz_offset = last_fuzz_loc - prep_buf;

	/* resize prep_buf */
	prep_buf_loc_offset = prep_buf_loc - prep_buf;
	prep_buf_len = prep_buf_len + fuzz_buf_len - (send_buf_loc - fuzz_tag_loc);
	prep_buf = tj_realloc(prep_buf, prep_buf_len);
	prep_buf_loc = prep_buf + prep_buf_loc_offset;

	/* restore last_fuzz_loc */
	if (last_fuzz_loc)
	  last_fuzz_loc = prep_buf + last_fuzz_offset;

	if (!do_last)
	  {
	  /* set random fuzzing type */
	  if (fuzz_type == r_random)
		fuzz_type = rand() % 9 + 1;

	  /* save last fuzz location */
	  last_fuzz_loc = prep_buf_loc;

	  /* insert characters */
	  for (i = 1; i <= fuzz_buf_len; i++)
		*prep_buf_loc++ = fuzz_char(fuzz_type);
	  }
	else
	  {
	  /* just insert last fuzz */
	  memcpy(prep_buf_loc, last_fuzz_loc, fuzz_buf_len);
	  prep_buf_loc += fuzz_buf_len;
	  }
	}

  if (tags_processed)
	{
	/* add last part */
	memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

	/* copy prep buf to send buf */
	send_buf_len = prep_buf_len;
	send_buf = tj_realloc(send_buf, send_buf_len);
	memcpy(send_buf, prep_buf, send_buf_len);
	}
  *send_buf_p = send_buf;
  done:

  free(prep_buf);
  return send_buf_len;
  }

int count_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len)
  {
  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc = prep_buf;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;

  char *send_buf = *send_buf_p;
  char *send_buf_loc = send_buf;
  char *count_tag = "<count ";
  char *count_tag_loc = NULL;
  char tag_data[128];
  char count_tag_type[16];
  char counter_text[16];
  int counter_text_len = 0;
  int tags_processed = 0;
  int i, vals;
  int count_tag_min, count_tag_max;
  static int *counters;
  int counter_num = 0;
  int init = 0;

  while ((count_tag_loc = memmem(send_buf_loc, (send_buf + send_buf_len
	  - send_buf_loc), count_tag, strlen(count_tag))))
	{
	count_tag_min = 0;
	count_tag_max = 0;

	/* set init mode */
	if (!counters)
	  {
	  init = 1;
	  inst_data->sess_data_modified = 0;
	  }
	else if (inst_data->gui == 1 && inst_data->sess_data_modified == 1)
	  {
	  init = 1;
	  if (counters)
		inst_data->sess_data_modified = 0;
	  }

	/* add text up to tag */
	memcpy(prep_buf_loc, send_buf_loc, count_tag_loc - send_buf_loc);

	/* advance prep_buf_loc to end */
	prep_buf_loc += (count_tag_loc - send_buf_loc);

	/* advance send_buf_loc to end of tag */
	send_buf_loc = count_tag_loc + strlen(count_tag);

	/* advance send_buf_loc to end of tag */
	if (!(send_buf_loc = memmem(send_buf_loc, send_buf + send_buf_len
		- send_buf_loc, ">", 1)))
	  goto done;
	send_buf_loc++;

	/* store tag data */
	memset(&tag_data, 0, 128);
	strncpy(tag_data, count_tag_loc + strlen(count_tag), send_buf_loc
		- count_tag_loc - strlen(count_tag));

	/* parse tag data */
	vals = sscanf(tag_data, "%s %d %d", count_tag_type, &count_tag_min,
		&count_tag_max);

	if ((vals < 3 && (strcmp(count_tag_type, "last")) == 0) || count_tag_min
		> count_tag_max)
	  {
	  tj_error(inst_data, "syntax error in <count> tag ", "", c_red);
	  goto done;
	  }

	if (strcmp(count_tag_type, "last") != 0)
	  {
	  /* setup counters */
	  if (init)
		{
		counter_num++;
		counters = realloc(counters, (counter_num + 1) * sizeof(int));
		if (counters == NULL)
		  {
		  printf("tcpjunk: memory allocation failed, terminating...\n");
		  exit(EXIT_FAILURE);
		  }
		if (inst_data->server_mode == 1)
		  counters[counter_num] = count_tag_min + (inst_data->inst_num - 1)
			  % (count_tag_max - count_tag_min + 1);
		else
		  counters[counter_num] = count_tag_min;
		}
	  else
	  /* increment counters */
		{
		counter_num++;
		if (counters[counter_num] == count_tag_max)
		  counters[counter_num] = count_tag_min;
		else
		  counters[counter_num]++;
		}

	  /* prepare digits */
	  if (strcmp(count_tag_type, "hex") == 0)
		sprintf(counter_text, "%x", counters[counter_num]);
	  else if (strcmp(count_tag_type, "dec") == 0)
		sprintf(counter_text, "%d", counters[counter_num]);
	  counter_text_len = strlen(counter_text);
	  }

	/* resize prep_buf */
	prep_buf_loc_offset = prep_buf_loc - prep_buf;
	prep_buf_len = prep_buf_len + counter_text_len - (send_buf_loc
		- count_tag_loc);
	prep_buf = tj_realloc(prep_buf, prep_buf_len);
	prep_buf_loc = prep_buf + prep_buf_loc_offset;

	/* insert digits */
	for (i = 0; i <= counter_text_len; i++)
	  {
	  *prep_buf_loc = counter_text[i];
	  prep_buf_loc++;
	  }
	prep_buf_loc--;

	tags_processed = 1;
	}

  if (tags_processed)
	{
	/* add last part */
	memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

	/* copy prep buf to send buf */
	send_buf_len = prep_buf_len;
	send_buf = tj_realloc(send_buf, send_buf_len);
	memcpy(send_buf, prep_buf, send_buf_len);
	}
  *send_buf_p = send_buf;
  done:

  if (init == 1)
	init = 0;

  free(prep_buf);
  return send_buf_len;

  }

int tamper_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len)
  {
  char *send_buf = *send_buf_p;
  char *send_buf_loc;
  char *tamp_tag = "<tamp>";
  char *tamp_tag_end = "</tamp>";
  char *tamp_tag_loc = NULL;
  char *tamp_text_loc = NULL;
  int tamp_text_len = 0;
  char tag_data[128];
  int tamp_min;
  int tamp_max;
  int tags_processed = 0;
  int vals;

  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;
  prep_buf_loc = prep_buf;
  send_buf_loc = send_buf;

  while ((tamp_tag_loc = memmem(send_buf_loc, (send_buf + send_buf_len
	  - send_buf_loc), tamp_tag, strlen(tamp_tag))))
	{
	tags_processed = 1;

	/* add text up to tag */
	memcpy(prep_buf_loc, send_buf_loc, tamp_tag_loc - send_buf_loc);

	/* advance prep_buf_loc to end */
	prep_buf_loc += (tamp_tag_loc - send_buf_loc);

	/* advance send_buf_loc to end of tag */
	send_buf_loc = tamp_tag_loc + strlen(tamp_tag);

	/* location of tamper text */
	tamp_text_loc = send_buf_loc;

	/* parse repeat tag data */
	vals = sscanf(tag_data, "%d %d", &tamp_min, &tamp_max);
	if (vals < 1)
	  goto end;
	if (vals == 1 && tamp_min < 1)
	  goto end;
	if (vals == 2 && (tamp_min < 1 || tamp_max < 1 || tamp_min > tamp_max))
	  goto end;

	/* advance send_buf_loc to end of tag data */
	if (!(send_buf_loc = memmem(send_buf_loc, (send_buf + send_buf_len
		- send_buf_loc), tamp_tag_end, strlen(tamp_tag_end))))
	  goto end;

	/* repeat text must exist before closing tag */
	if (tamp_text_loc > send_buf_loc)
	  goto end;

	/* calculate repeat text length */
	tamp_text_len = send_buf_loc - tamp_text_loc;

	/* advance send_buf_loc to end of end_tag */
	send_buf_loc += strlen(tamp_tag_end);

	/* resize prep_buf */
	prep_buf_loc_offset = prep_buf_loc - prep_buf;
	prep_buf_len = prep_buf_len + tamp_text_len - (send_buf_loc - tamp_tag_loc);
	prep_buf = tj_realloc(prep_buf, prep_buf_len);
	prep_buf_loc = prep_buf + prep_buf_loc_offset;

	/* start repetition */
	memcpy(prep_buf_loc, tamp_text_loc, tamp_text_len);
	prep_buf_loc += tamp_text_len;

	}

  if (tags_processed)
	{

	/* add last part from send buf to prep buf */
	memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

	/* copy prep buf to send buf */
	send_buf_len = prep_buf_len;
	send_buf = tj_realloc(send_buf, send_buf_len);
	memcpy(send_buf, prep_buf, send_buf_len);
	}

  end:

  free(prep_buf);
  *send_buf_p = send_buf;
  return send_buf_len;
  }

int encode_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len)
  {
  char *send_buf = *send_buf_p;
  char *send_buf_loc;
  char *enc_tag = "<enc ";
  char *enc_tag_end = "</enc>";
  char *enc_tag_loc = NULL;
  char *enc_text_loc = NULL;
  int enc_text_len = 0;
  char tag_data[128];
  int tags_processed;
  int vals;
  char dig_type[16];
  char enc_pre[16];
  char enc_post[16];
  int enc_pad;
  int enc_type;

  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;
  char *enced_buf = NULL;

  tags_processed = 0;
  prep_buf_loc = prep_buf;
  send_buf_loc = send_buf;

  while ((enc_tag_loc = memmem(send_buf_loc, (send_buf + send_buf_len
	  - send_buf_loc), enc_tag, strlen(enc_tag))))
	{
	tags_processed = 1;

	/* add text up to tag */
	memcpy(prep_buf_loc, send_buf_loc, enc_tag_loc - send_buf_loc);

	/* advance prep_buf_loc to end */
	prep_buf_loc += (enc_tag_loc - send_buf_loc);

	/* advance send_buf_loc to end of tag */
	send_buf_loc = enc_tag_loc + strlen(enc_tag);

	/* location of text to encode */
	if (!(enc_text_loc = memmem(send_buf_loc, send_buf + send_buf_len
		- send_buf_loc, ">", 1)))
	  goto end;
	enc_text_loc++;

	/* store tag data */
	memset(&tag_data, 0, 128);
	strncpy(tag_data, enc_tag_loc + strlen(enc_tag), enc_text_loc - enc_tag_loc
		- strlen(enc_tag));

	/* parse tag data */
	enc_pad = 0;
	vals = sscanf(tag_data, "%s%s%s%d", enc_pre, dig_type, enc_post, &enc_pad);

	if (vals < 4 || enc_pad > 15)
	  {
	  tj_error(inst_data, "syntax error in <enc> tag ", "", c_red);
	  goto end;
	  }

	/* get digit type */
	enc_type = get_dig_type(dig_type, inst_data);
	if (enc_type == 0)
	  goto end;

	/* advance send_buf_loc to end of tag data */
	if (!(send_buf_loc = memmem(send_buf_loc, (send_buf + send_buf_len
		- send_buf_loc), enc_tag_end, strlen(enc_tag_end))))
	  goto end;

	/* tag text must exist before closing tag */
	if (enc_text_loc > send_buf_loc)
	  goto end;

	/* calculate tag text length */
	enc_text_len = send_buf_loc - enc_text_loc;
	if (enc_text_len == 0)
	  goto end;

	/* advance send_buf_loc to end of end_tag */
	send_buf_loc += strlen(enc_tag_end);

	/* encode */
	enced_buf = tj_malloc(enced_buf, enc_text_len * (12 + enc_pad));
	encode(enced_buf, enc_text_loc, enc_text_len, enc_pre, enc_type, enc_post,
		enc_pad);

	/* resize prep_buf */
	prep_buf_loc_offset = prep_buf_loc - prep_buf;
	prep_buf_len = prep_buf_len + strlen(enced_buf) - (send_buf_loc
		- enc_tag_loc);
	prep_buf = tj_realloc(prep_buf, prep_buf_len);
	prep_buf_loc = prep_buf + prep_buf_loc_offset;

	/* insert encoded text */
	memcpy(prep_buf_loc, enced_buf, strlen(enced_buf));
	prep_buf_loc += strlen(enced_buf);
	free(enced_buf);

	}

  if (tags_processed)
	{

	/* add last part from send buf to prep buf */
	memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

	/* copy prep buf to send buf */
	send_buf_len = prep_buf_len;
	send_buf = tj_realloc(send_buf, send_buf_len);
	memcpy(send_buf, prep_buf, send_buf_len);
	}

  end:

  free(prep_buf);
  *send_buf_p = send_buf;
  return send_buf_len;
  }
