/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <openssl/crypto.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#define VER "2.7.01"
#define SOCK_BUF_LEN 16384
#define BACKLOG 10
#define MAX_SECTS 100
#define SERVER_CERT "server.crt"
#define SERVER_KEY "server.key"
#define SECT_START "<section>"
#define SECT_END "</section>"

enum error_colors
  {
  c_black,
  c_red,
  c_red_hl,
  c_blue_hl
  };

enum auto_file_mode
  {
  f_none,
  f_rand,
  f_seq
  };

struct tj_data
  {
    char *dest_addr;
    int dest_addr_len;
    int dest_port;
    char *src_addr;
    int src_addr_len;
    int src_port;
    char src_ip[32];
    char interface[32];
    int offset;
    int rand_ip;

    char *sess_data;
    int sess_data_len;
    char sess_file_name[FILENAME_MAX];
    int sess_data_modified;
    char *sects[MAX_SECTS];
    int sects_len[MAX_SECTS];
    int sect_idx;
    int auto_file;
    char **files_list;
    int auto_files_num;

    int server_mode;
    int gui;
    int debug;
    int verbose;
    int hexdump;
    int wait;
    int conn_timeout;
    int reply_timeout;
    unsigned int keepalive;
    unsigned int loop;
    int srv_sock;
    int wait_conn;
    int no_wait_last;
    int use_ssl;
    char server_cert[32];
    char server_key[32];
    int seg_size;
    int seg_delay;
    unsigned int inst_num;
  };

/* junk.c */
void junk(struct tj_data *inst_data);

/* sjunk.c */
void sjunk(struct tj_data *inst_data);

/* utils.c */
int tj_connect(int sock, struct sockaddr_in sa, struct tj_data *inst_data);
void tj_disconnect(int sock, struct tj_data *inst_data, SSL * ssl,
    SSL_CTX * ctx);
int read_reply(int sock, SSL *ssl, struct tj_data *inst_data);
void send_data(int sock, SSL *ssl, struct tj_data *inst_data, char *send_buf,
    int send_buf_len);
char fuzz_char(int ftype);
void sigchld_handler(int s);
int check_address(struct tj_data *inst_data, char *addr);
int check_src_address(struct tj_data *inst_data, char *addr);
void load_session_data(struct tj_data *inst_data);
long tj_filesize(FILE *stream);
int check_port(int port, struct tj_data *inst_data);
void load_file_list(struct tj_data *inst_data, char *rand_file_dir);
void set_session_file(struct tj_data *inst_data);
void find_sections(struct tj_data *inst_data);
void tj_error(struct tj_data *inst_data, char *message, char *param, int color);
void
    set_src_ip(struct tj_data *inst_data, char *bind_if_name, char *new_src_ip);
void tj_show_help();
char*tj_malloc(char *buf, int buf_len);
char *tj_realloc(char *buf, int buf_len);
void cleanup(int sig);
char *hex_dump(char *buf, int buf_len);

/* gui.c */
void gui_init(int argc, char *argv[], struct tj_data *gui_inst_data);
void gui_show_traffic(char *buf, int buf_len, int color,
    struct tj_data *gui_inst_data);
void gui_add_log(char *buf, int color);
void gui_accept(struct tj_data *gui_inst_data);

/* tags.c */
int fuzz_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len);
int system_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len);
int repeat_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len);
int encode_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len);
int
    instance_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len);
int char_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len);
int count_tag(char **send_buf_p, struct tj_data *inst_data, int send_buf_len);

